from typing import List, Dict, Union

from sqlalchemy.engine import Result
from sqlalchemy.orm import Session
from sqlalchemy.engine import Engine
from sqlalchemy import delete

from texttable import Texttable

from .models import Base

COL_WIDTH_SPACER = 2
MAX_COL_WIDTH = 30


def db_merge(engine: Engine, entry: Base):
    """
    Merges an ORM object
    """
    with Session(engine) as session:
        session.merge(entry)
        session.commit()


def result_to_cli_string(result: Result) -> str:
    """
    Returns a string that prints out a nice table of sqlalchemy results
    on the cli
    """

    # width of each column
    col_widths = {}

    # list of dicts denoting the table to output
    output = []

    # first we're size our column widths based on the length of the data
    for row in result.scalars():
        # initializing column widths to match the header column + spacer
        if not col_widths:
            for col in row.__table__.columns:
                col_widths[col.name] = len(col.name) + COL_WIDTH_SPACER

        for col_name, col_width in col_widths.items():
            row_col_width = len(str(getattr(row, col_name, "")))
            if row_col_width + COL_WIDTH_SPACER > col_width:
                col_widths[col_name] = row_col_width + COL_WIDTH_SPACER

        # saving row output into our list - you can only iterate over `results.scalars` once!
        output.append({c: str(getattr(row, c, "")) for c in col_widths})

    # now we will turn our output list of lists into a giant string sized
    # on our column widths
    output_str = "".join(
        [c_name.ljust(c_width) for c_name, c_width in col_widths.items()]
    )
    for row in output:
        output_str += "\n" + "".join(
            [row[c_name].ljust(c_width) for c_name, c_width in col_widths.items()]
        )

    return output_str


def delete_and_rebuild_device(
    session: Session, device: str, table: Base, data: Union[Dict, List[Dict]]
):
    """
    Takes in an existing db session, the name of a table, name of the device,
    and a list of rows to update.
    First removes all rows matching the key {col_name, col_data},
    Then inserts ("adds") rows in the data.
    """
    session.execute(delete(table).where(getattr(table, "device") == device))
    session.commit()

    if isinstance(data, dict):
        data = [data]

    for row in data:
        entry = table()
        try:
            setattr(entry, "device", device)
            for k, v in row.items():
                setattr(entry, k, v)
        except AttributeError:
            print(f"Dict to db error for {table}: {row}")
            del entry
            continue

        session.add(entry)
    session.commit()



# pylint: disable=too-many-ancestors, missing-class-docstring, too-few-public-methods
from datetime import timedelta, datetime
from typing import Optional, List, Any
from json import dumps


from netaddr import *

from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.orm import DeclarativeBase
import sqlalchemy.types as types
from sqlalchemy.dialects import postgresql
from sqlalchemy import String, ARRAY

from postgresql_audit import versioning_manager

__all__ = (
    "Base",
    "Device",
    "Neighbor",
    "Lag",
    "ARP",
    "IPInterface",
    "Route",
    "VNI",
    "MPLS",
    "Inventory",
)


class DBModelError(Exception):
    pass


##########################################
#### SQLalchemy special Network Types ####
# Note that we're using the netaddr library because
# its classes are for both IPv4 and IPv6
class IPAddressType(types.TypeDecorator):
    impl = postgresql.INET
    cache_ok = True

    def process_bind_param(self, value, dialect):
        return str(value) if value else None

    def process_result_value(self, value, dialect):
        return IPAddress(str(value)) if value else None

    @property
    def python_type(self):
        return IPAddress


class IPNetworkType(types.TypeDecorator):
    impl = postgresql.INET
    cache_ok = True

    def process_bind_param(self, value, dialect):
        return str(value) if value else None

    def process_result_value(self, value, dialect):
        return IPNetwork(str(value)) if value else None

    @property
    def python_type(self):
        return IPNetwork


class MACAddressType(types.TypeDecorator):
    impl = postgresql.MACADDR
    cache_ok = True

    def process_bind_param(self, value, dialect):
        return str(value) if value else None

    def process_result_value(self, value, dialect):
        return EUI(value) if value else None

    @property
    def python_type(self):
        return EUI


#####################################################
#### SQLalchemy Base models and versioning manager ####
class Base(DeclarativeBase):

    def as_dict(self):
        """
        Return rows as dict
        """
        return {c.name: getattr(self, c.name) for c in self.__table__.c}

    def as_json(self):
        """
        Return rows as json
        """
        return dumps(self.as_dict())

    @classmethod
    def from_dict(cls, input: dict):
        """
        Builds a row object from an inputted dictionary, casting from strings/
        integers/etc to the appropriate "mapped" python types for our models
        """
        casted = {}
        for col_name, value in input.items():

            if col_name not in cls.__table__.c.keys():
                raise DBModelError(
                    f"Invalid column name {col_name} for {cls.__tablename__}"
                )

            col = cls.__table__.c[col_name]

            # for list types, evaluate each item
            if col.type._is_array:
                if not isinstance(value, list):
                    raise DBModelError(
                        f"column {col_name} for {cls.__tablename__} requires list input"
                    )
                casted[col_name] = [cls._cast_value(i, col) for i in value]

            # for non-list types, convert input value
            else:
                casted[col_name] = cls._cast_value(value, col)

        return cls(**casted)

    @classmethod
    def _cast_value(cls, value, col):
        """
        Casts a particular input value to the db type
        """

        # forcing empty strings/lists etc to 'None'
        # for nullable values
        if col.nullable and not value:
            return None

        try:
            return col.type.python_type(value)
        except Exception as e:
            raise DBModelError(
                f"Invalid input '{value}' for {cls.__tablename__} {col.name}: {e}"
            )

    def __repr__(self):
        """
        Row repr, eg "Table(pk1_name='pk1_val', pk2_name='pk2_val'..)"
        """
        return (
            f"{self.__class__.__name__}("
            + ", ".join(
                [
                    f"{k.name}='{getattr(self, k.name)}'"
                    for k in self.__table__.primary_key.columns
                ]
            )
            + ")"
        )


versioning_manager.init(Base)


# mapped columns must be defined after versioning manager is initialized
class Versioned(object):
    """
    Versioned mixin
    """

    __versioned__ = {"exclude": ["last_updated"]}
    first_seen: Mapped[datetime]
    last_updated: Mapped[datetime]


class NonVersioned(object):
    """
    NonVersioned Mixin
    """

    first_seen: Mapped[datetime]
    last_updated: Mapped[datetime]


####################################
#### umnetdb table definitions ####
class Device(Versioned, Base):
    __tablename__ = "device"

    __versioned__ = {"exclude": ["last_updated", "uptime"]}

    name: Mapped[str] = mapped_column(primary_key=True)
    ip: Mapped[IPAddress] = mapped_column(IPAddressType)
    version: Mapped[Optional[str]]
    vendor: Mapped[Optional[str]]
    model: Mapped[Optional[str]] = mapped_column(String(40))
    serial: Mapped[Optional[str]]
    uptime: Mapped[timedelta]


class Neighbor(Versioned, Base):
    __tablename__ = "neighbor"

    device: Mapped[str] = mapped_column(primary_key=True)
    port: Mapped[str] = mapped_column(primary_key=True)
    remote_device: Mapped[str]
    remote_port: Mapped[str]


class Lag(Versioned, Base):
    """
    Table of LAG members, statuses, and their parent mappings
    """

    __tablename__ = "lag"

    device: Mapped[str] = mapped_column(primary_key=True)
    member: Mapped[str] = mapped_column(primary_key=True)
    parent: Mapped[str] = mapped_column(primary_key=True)
    protocol: Mapped[str]
    admin_up: Mapped[bool]
    oper_up: Mapped[bool]
    peer_link: Mapped[bool]
    vpc_id: Mapped[int]


class ARP(NonVersioned, Base):
    __tablename__ = "arp"

    device: Mapped[str] = mapped_column(primary_key=True)
    interface: Mapped[str] = mapped_column(String(50), primary_key=True)
    ip: Mapped[IPAddress] = mapped_column(IPAddressType, primary_key=True)
    mac: Mapped[EUI] = mapped_column(MACAddressType, primary_key=True)


class IPInterface(Versioned, Base):
    __tablename__ = "ip_interface"

    device: Mapped[str] = mapped_column(primary_key=True)
    ip_address: Mapped[IPNetwork] = mapped_column(IPNetworkType, primary_key=True)
    interface: Mapped[str] = mapped_column(String(50), primary_key=True)
    description: Mapped[Optional[str]]
    mtu: Mapped[int]
    admin_up: Mapped[bool]
    oper_up: Mapped[bool]
    vrf: Mapped[Optional[str]]
    secondary: Mapped[bool]
    helpers: Mapped[list[str]] = mapped_column(ARRAY(String))


class Route(Versioned, Base):
    """
    Device route tables. Note that we mean *active* routes (the FIB).
    In reality pulling the FIB directly off a device is not
    straightforward, and helpful information like protocol and age are not present.
    As a result, this data is actually populated via "show ip route vrf all" (cisco),
    "show route active-path" (junos), "show routing route" (panos)
    """

    __tablename__ = "route"

    # also want to exclude 'age' from being versioned
    __versioned__ = {"exclude": ["last_updated", "age"]}

    device: Mapped[str] = mapped_column(primary_key=True)
    vrf: Mapped[str] = mapped_column(primary_key=True)
    prefix: Mapped[IPNetwork] = mapped_column(IPNetworkType, primary_key=True)
    learned_from: Mapped[str] = mapped_column(primary_key=True)

    nh_ip: Mapped[Optional[IPAddress]] = mapped_column(IPAddressType)
    nh_interface: Mapped[Optional[str]]
    nh_table: Mapped[str]

    protocol: Mapped[str]
    age: Mapped[Optional[timedelta]]

    mpls_label: Mapped[Optional[list[str]]] = mapped_column(ARRAY(String))
    vxlan_vni: Mapped[Optional[int]]
    vxlan_endpoint: Mapped[Optional[IPAddress]] = mapped_column(IPAddressType)


class VNI(Versioned, Base):
    __tablename__ = "vni"

    device: Mapped[str] = mapped_column(primary_key=True)
    vni: Mapped[str] = mapped_column(primary_key=True)
    vrf: Mapped[Optional[str]]
    vlan_id: Mapped[Optional[int]]


class MPLS(Versioned, Base):
    """
    MPLS forwarding tables.
    Note that for aggregate labels the "nh_interface" is actually the
    VRF name!
    """

    __tablename__ = "mpls"

    device: Mapped[str] = mapped_column(primary_key=True)
    in_label: Mapped[int] = mapped_column(primary_key=True)
    out_label: Mapped[list[str]] = mapped_column(ARRAY(String))
    nh_interface: Mapped[str] = mapped_column(primary_key=True)
    aggregate: Mapped[bool]

    # NB: string instead of IP type because I've seen malformed FEC data
    # (eg 2607:f018:fffd:100:ffff:ffff:ffff:ff48/12)
    fec: Mapped[Optional[str]]
    nh_ip: Mapped[Optional[IPAddress]] = mapped_column(IPAddressType)
    rd: Mapped[Optional[str]]


class Inventory(Versioned, Base):
    """
    Inventory data
    """

    __tablename__ = "inventory"

    device: Mapped[str] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(primary_key=True)
    type: Mapped[str] = mapped_column(primary_key=True)
    part_number: Mapped[Optional[str]]
    serial_number: Mapped[Optional[str]]



from ipaddress import ip_address, ip_network
from typing import List

from sqlalchemy import select, text
from sqlalchemy.orm import Session

from .models import Route, IPInterface, Neighbor, MPLS, VNI


class DBLookupError(Exception):
    """ "
    Generic db query error
    """

    def __init__(self, query, table):
        super().__init__(f"Could not find {query} in {table.__name__} table")


def neighbor_query(session: Session, device: str, port: str) -> Neighbor:
    """
    Looks up the neighbor object for a particular device in a particular row
    """
    result = session.scalar(
        select(Neighbor).where(Neighbor.device == device).where(Neighbor.port == port)
    )
    if not result:
        raise DBLookupError(f"{device} {port}", Neighbor)

    return result


def ip_interface_query(session: Session, ip: ip_address) -> IPInterface:
    """
    Finds the router that hosts this IP address (its default gateway)
    by querying the IPInterface table
    """
    result = session.scalar(
        select(IPInterface).where(text(f"'{ip}' = host(ip_address)"))
    )
    if not (result):
        raise DBLookupError(ip, IPInterface)
    return result


def lpm_query(s: Session, router: str, vrf: str, dest_ip: str) -> List[Route]:
    """
    Get the lpm(s) for an IP address on a particular router in a particular VRF
    """
    query = (
        select(Route)
        .where(text(f"'{dest_ip}' <<= prefix"))
        .where(Route.device == router)
        .where(Route.vrf == vrf)
        .order_by(text("masklen(prefix) desc"))
    )
    routes = s.scalars(query)

    if not routes:
        raise DBLookupError(f"{router} {vrf} {dest_ip}", Route)

    # Routes are ordered by prefix length - the longest ones
    # are first.
    lpm = None
    lpm_routes = []
    for route in routes:
        # if lpm is none this is the first loop iteration - set this route
        # to be the LPM and continue
        if lpm is None:
            lpm = ip_network(route.prefix).prefixlen
            lpm_routes.append(route)
            continue

        # if this route's prefixlen matches the lpm, save it to
        # our list of lpms
        if ip_network(route.prefix).prefixlen == lpm:
            lpm_routes.append(route)

        # otherwise the route has a shorter lpm and we've found them all
        elif ip_network(route.prefix).prefixlen <= lpm:
            break

    return lpm_routes


def mpls_query(s: Session, router: str, label: str) -> List[MPLS]:
    """
    Looks up a particular label on a particular router.
    Returns the MPLS label entries
    """
    query = select(MPLS).where(MPLS.device == router).where(MPLS.in_label == label)
    results = s.scalars(query)

    if not results:
        raise DBLookupError(f"{router} {label}", MPLS)

    return results


def l3vni_query(s: Session, router: str, vni: int) -> str:
    """
    Looks up a VNI on the router and returns the
    appropriate VRF
    """
    query = select(VNI.vrf).where(VNI.device == router).where(VNI.vni == vni)
    results = s.scalar(query)

    if not results:
        raise DBLookupError(f"vrf from {router} {vni}", VNI)

    return results
